from dataclasses import dataclass

from . import entities as en


@dataclass
class Rship:
    source: en.Entity
    target: en.Entity


@dataclass
class TableToTable(Rship):
    source: en.Table
    target: en.Table



import argparse
import logging
import os

from .parsers import QueryParser, YamlParser
from .helpers import print_basics

parser = argparse.ArgumentParser(description='Extract info from your SQLs')
parser.add_argument('command', help='command to execute', type=str, nargs=1)
parser.add_argument('-f', '--file', help='SQL/YAML file')
parser.add_argument('-l', '--loglevel', help='logging level', default='INFO')
args = parser.parse_args()

fmt = '%(asctime)s | %(levelname)s | %(module)s | %(funcName)s | %(message)s'
lvl = getattr(logging, args.loglevel.upper())
logging.basicConfig(format=fmt, level=lvl)


def get_parser(file: str, target: str = None) -> QueryParser:
    _, ext = os.path.splitext(file)
    assert ext == '.sql', 'Not SQL file'
    query = open(file, 'r').read()
    parser = QueryParser(query, target=target)
    return parser


def main():

    command = args.command[0]

    if command == 'bulk':
        _, ext = os.path.splitext(args.file)
        assert ext == '.yaml', 'Input file not yaml'
        bulk = YamlParser(args.file)
        bulk.parse_jobs()
        bulk.save_output()
        return

    if command == 'structure':
        parser = get_parser(args.file)
        print_basics(parser.basics)
        return

    if command == 'sources':
        parser = get_parser(args.file)
        for table in parser.extract_sources():
            print(table)
        return

    if command == 'webserver':
        from .web import app
        app.run(debug=True)

    raise Exception(f'Unknown command: {command}')


if __name__ == '__main__':
    main()




import copy
import glob
import json
import os

import yaml
from flask import render_template, redirect, request, url_for

from sqlexpress.parsers import YamlParser
from sqlexpress.web import app
from sqlexpress.web.utils import (
    extract_project,
    extract_nodes_edges_from_output,
)


FOLDER = os.environ.get('WEBSERVER_FOLDER')

PROJECTS = {
    p['filename']: p for p in [
        extract_project(f) for f in sorted(glob.glob(f'{FOLDER}/*.yaml'))
    ]
}


@app.route('/')
def home():
    projects = []
    for filename, project in PROJECTS.items():
        tmp = {'name': project['name'], 'filename': filename}
        if project['output']:
            nodes, _ = extract_nodes_edges_from_output(project['output'])
            tmp['n_jobs'] = len(project['output']['jobs'])
            tmp['n_tables'] = len(nodes)
        projects.append(tmp)
    return render_template('home.html', projects=projects)


@app.route('/project/<string:filename>/extract/', methods=['POST'])
def project_extract(filename: str):
    filepath = PROJECTS[filename]['filepath']
    bulk = YamlParser(filepath)
    bulk.parse_jobs()
    bulk.save_output()
    PROJECTS[filename] = extract_project(filepath)
    return redirect(url_for('home'))


@app.route('/project/<string:filename>/input/', methods=['GET'])
def project_input(filename: str):
    p = copy.deepcopy(PROJECTS[filename])
    p['input'] = yaml.safe_dump(p['input'], sort_keys=False)
    p['input'] = p['input'].replace('\njobs', '\n\njobs')
    p['input'] = p['input'].replace('\n- ', '\n\n- ')
    return render_template(f'project_input.html', project=p)


@app.route('/project/<string:filename>/output/', methods=['GET'])
def project_output(filename: str):
    p = copy.deepcopy(PROJECTS[filename])
    p['output'] = json.dumps(p['output'], indent=4)
    return render_template(f'project_output.html', project=p)


@app.route('/project/<string:filename>/graph/', methods=['GET', 'POST'])
def project_graph(filename: str):
    p = copy.deepcopy(PROJECTS[filename])
    q = request.form.get('query', '')
    p['nodes'], p['edges'] = extract_nodes_edges_from_output(p['output'], q)
    return render_template(f'project_graph.html', project=p, query=q)


@app.route('/project/<string:filename>/tables/', methods=['GET', 'POST'])
def project_tables(filename: str):
    p = copy.deepcopy(PROJECTS[filename])
    p['nodes'], _ = extract_nodes_edges_from_output(p['output'])
    q = request.form.get('query', '')
    if q != '':
        p['nodes'] = [n for n in p['nodes'] if q in n]
    return render_template(f'project_tables.html', project=p, query=q)



import copy
import re
from dataclasses import dataclass, field
from typing import List

import networkx as nx

from .. import helpers as hp
from ..structures import clauses as cl, basics as bs

CLAUSE_LIST = [
    cl.SelectClause(),
    cl.FromClause(),
    cl.WhereClause(),
    cl.LeftJoinClause(),
    cl.RightJoinClause(),
    cl.FullJoinClause(),
    cl.InnerJoinClause(),
    cl.JoinClause(),
    cl.OnClause(),
    cl.GroupByClause(),
    cl.OrderByClause(),
    cl.HavingClause(),
    cl.UnionClause(),
    cl.CreateTempFuncClause(),
    cl.WithClause(),
    cl.CteStartClause(),
    cl.CteEndClause(),
    cl.SubqueryStartClause(),
    cl.SubqueryEndClause(),
]


@dataclass
class QueryParser:
    raw: str
    target: str = None
    env: dict = None

    processed: str = field(init=False)
    clauses: List[cl.Clause] = field(init=False)
    basics: List[bs.BasicQuery] = field(init=False)
    G_table: nx.DiGraph = field(init=False)

    def __post_init__(self) -> None:
        self.preprocess()
        self.parse_clauses()
        self.parse_basics()
        self.parse_table_graph()

    def preprocess(self) -> None:
        processed = self.raw
        if self.env:
            for key, value in self.env.items():
                processed = processed.replace(f'{{{key}}}', value)
        processed = hp.remove_comments(processed)
        replace = [
            (r'\(', ' ( '),  # add blanks before & after open bracket
            (r'\)', ' ) '),  # add blanks before & after close bracket
            (r'\,', ' , '),  # add blanks before & after comma
            (r'\n', ' '),  # remove next line
            (r' +', ' '),  # remove multiple blanks
        ]
        for frm, to in replace:
            processed = re.sub(frm, to, processed)
        self.processed = processed

    def parse_clauses(self) -> None:
        self.clauses: List[cl.Clause] = []
        tokens: List[str] = self.processed.split(' ')
        n_tokens: int = len(tokens)

        # loop through tokens
        for i, token in enumerate(tokens):

            # to speed up for loop in next step
            shortlisted_clauses = [
                c for c in CLAUSE_LIST
                if c.match_first_pattern(token)
            ]

            # confirm if start of new clause
            for cls in shortlisted_clauses:

                # check if clause's patterns & pre_patterns match tokens
                is_pattern_match = False
                upper_idx = i+cls.n_patterns
                lower_idx = i-cls.n_pre_patterns
                if upper_idx <= n_tokens and lower_idx >= 0:
                    is_pattern_match = cls.match_patterns(
                        tokens=tokens[i: upper_idx],
                        pre_tokens=tokens[lower_idx: i],
                    )

                # check if previous clause is expecting token
                is_pending_token = False if len(self.clauses) == 0 \
                    else self.clauses[-1].pending_token()

                # if patterns match & not expecting token, add new clause
                if is_pattern_match and not is_pending_token:
                    self.clauses.append(copy.deepcopy(cls))
                    break

            # add current token to most recent clause
            self.clauses[-1].add_token(token)

    def parse_basics(self) -> None:
        self.basics: List[bs.BasicQuery] = []

        # loop through clauses
        for i, clause in enumerate(self.clauses):
            is_pending = False if len(self.basics) == 0 \
                else self.basics[-1].pending_clause()

            # start of new clause
            if isinstance(clause, cl.CteStartClause):
                self.basics.append(bs.CteBasicQuery())
            if isinstance(clause, cl.SelectClause) and not is_pending:
                self.basics.append(bs.BasicQuery())

            # add clause to last basic query
            if len(self.basics) > 0:
                self.basics[-1].add_clause(clause)

        assert isinstance(self.basics[-1], bs.BasicQuery)
        self.basics[-1].target = self.target

    def parse_table_graph(self) -> None:
        rships = [r for b in self.basics for r in b.extract_table_to_table()]
        names = [(r.source.name, r.target.name) for r in rships]
        self.G_table = nx.DiGraph()
        self.G_table.add_edges_from(names)

    def extract_sources(self) -> List[str]:
        nodes = self.G_table.nodes
        return [n for n in nodes if self.G_table.in_degree(n) == 0]

    def print(self) -> None:
        hp.print_basics(self.basics)



import re
from abc import ABC
from typing import Any, List, Optional

from ..graph import entities as en


class Clause(ABC):

    def __init__(
            self, patterns: List[Any],
            pre_patterns: List[Any] = None,
            require_min_tokens: Optional[int] = None,
            require_balanced_brackets: bool = True,
    ) -> None:

        self.patterns = patterns
        self.n_patterns = len(self.patterns)

        self.pre_patterns = pre_patterns or []
        self.n_pre_patterns = len(self.pre_patterns)

        self.require_min_tokens = require_min_tokens or len(self.patterns)
        self.require_balanced_brackets = require_balanced_brackets

        self.tokens = []

    @staticmethod
    def _match_pattern_token(pattern: Any, token: str) -> bool:
        """ Check if pattern & token match """
        p_type = type(pattern)
        if p_type == str and pattern.lower() == token.lower():
            return True
        if p_type == re.Pattern and pattern.match(token.lower()):
            return True
        return False

    def match_first_pattern(self, token: str) -> bool:
        """ Check if token match 1st pattern """
        return self._match_pattern_token(self.patterns[0], token)

    def match_patterns(
            self, tokens: List[str],
            pre_tokens: List[str] = None,
    ) -> bool:
        """ Check if tokens match all patterns """
        pre_tokens = pre_tokens or []

        # flag if token count mismatch
        assert self.n_patterns == len(tokens)
        assert self.n_pre_patterns == len(pre_tokens)

        # false if any tokens do not match
        patterns = self.pre_patterns + self.patterns
        tokens = pre_tokens + tokens
        matches = [
            self._match_pattern_token(p, t)
            for p, t in zip(patterns, tokens)
        ]
        return all(matches)

    def add_token(self, token: str) -> None:
        """ Add token to existing list of tokens """
        self.tokens.append(token)

    def pending_token(self) -> bool:
        """ Check if clause is expecting more tokens """

        if self.require_balanced_brackets:
            n_open = sum([t == '(' for t in self.tokens])
            n_close = sum([t == ')' for t in self.tokens])
            if n_open > n_close:
                return True

        if self.require_min_tokens:
            if len(self.tokens) < self.require_min_tokens:
                return True

        return False

    def extract_tables(self) -> List[en.Table]:
        """ Depending on clause, may need to extract tables """
        return []

    def extract_target(self) -> Optional[str]:
        """ Only has non-None output for CteStartClause """
        return None

    def __repr__(self) -> str:
        """ If tokens too long, cut at 50 characters """
        raw = " ".join(self.tokens)
        raw = raw[:50] + '...' if len(raw) >= 50 else raw
        return f'{self.__class__.__name__}(raw=[{raw}])'


###############################################################################


class SelectClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['select'])


class FromClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['from'])

    def extract_tables(self) -> List[en.Table]:
        return [] if len(self.tokens) <= self.n_patterns \
            else [en.Table(name=self.tokens[self.n_patterns])]


class WhereClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['where'])


class JoinClause(Clause):
    def __init__(self, patterns: List[Any] = None) -> None:
        super().__init__(patterns=patterns or ['join'])

    def extract_tables(self) -> List[en.Table]:
        return [] if len(self.tokens) <= self.n_patterns \
            else [en.Table(name=self.tokens[self.n_patterns])]


class LeftJoinClause(JoinClause):
    def __init__(self) -> None:
        super().__init__(patterns=['left', 'join'])


class RightJoinClause(JoinClause):
    def __init__(self) -> None:
        super().__init__(patterns=['right', 'join'])


class FullJoinClause(JoinClause):
    def __init__(self) -> None:
        super().__init__(patterns=['full', 'join'])


class InnerJoinClause(JoinClause):
    def __init__(self) -> None:
        super().__init__(patterns=['inner', 'join'])


class OnClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['on'])


class GroupByClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['group', 'by'])


class OrderByClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['order', 'by'])


class HavingClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['having'])


class UnionClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['union'])


class CreateTempFuncClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['create', 'temp', 'function'])


class WithClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=['with'])


class CteStartClause(Clause):
    def __init__(self) -> None:
        super().__init__(
            patterns=[re.compile('([a-zA-Z0-9]+)'), 'as', '(', 'select'],
            require_min_tokens=3,
            require_balanced_brackets=False,
        )

    def extract_target(self) -> str:
        return self.tokens[0]


class CteEndClause(Clause):
    def __init__(self) -> None:
        super().__init__(
            patterns=[')', re.compile('(,|select){1}')],
            require_min_tokens=1,
        )


class SubqueryStartClause(Clause):
    def __init__(self) -> None:
        super().__init__(
            pre_patterns=[re.compile('(from|join|in){1}')],
            patterns=['(', 'select'],
            require_min_tokens=1,
            require_balanced_brackets=False,
        )


class SubqueryEndClause(Clause):
    def __init__(self) -> None:
        super().__init__(patterns=[')'])



import json
import os
from typing import Tuple

from sqlexpress.parsers import YamlParser


def extract_project(filepath: str) -> dict:
    filename = os.path.basename(filepath)
    yaml = YamlParser(filepath=filepath)
    try:
        output = json.load(open(yaml.params.output, 'r'))
    except:
        output = None
    project = {
        'name': os.path.splitext(filename)[0],
        'filepath': filepath,
        'filename': filename,
        'input': yaml.raw,
        'output': output,
    }
    return project


def extract_nodes_edges_from_output(output: dict, query: str = None) -> Tuple[list, list]:

    edges = []
    for job in output['jobs']:
        edges += [[src, job['target']] for src in job['sources']]

    if query is not None and query != '':
        edges = [e for e in edges if query in e[0] or query in e[1]]

    nodes = []
    for edge in edges:
        nodes += edge
    nodes = list(sorted(set(nodes)))

    return nodes, edges


import json
import logging
import os
from dataclasses import dataclass, field
from typing import List

import yaml

from . import QueryParser


@dataclass
class Params:
    output: str
    folder: str = None
    env: dict = None

    def to_dict(self) -> dict:
        output = {'output': self.output}
        if self.folder:
            output['folder'] = self.folder
        if self.env:
            output['env'] = self.env
        return output


@dataclass
class Job:
    target: str
    sql: str
    env: dict = None
    parser: QueryParser = None

    def parse(self, folder: str, env: dict = None) -> None:
        fullpath = os.path.join(folder, self.sql) if folder else self.sql
        _, ext = os.path.splitext(fullpath)
        assert ext == '.sql', f'Not a SQL file: {self.sql}'
        query = open(fullpath, 'r').read()
        combined_env = {**self.env, **env} if self.env and env \
            else self.env or env
        self.parser = QueryParser(query, env=combined_env)

    def to_dict(self) -> dict:
        output = {'target': self.target, 'sql': self.sql}
        if self.env:
            output['env'] = self.env
        if self.parser:
            output['sources'] = self.parser.extract_sources()
        return output


@dataclass
class YamlParser:
    filepath: str
    raw: dict = field(init=False)
    params: Params = field(init=False)
    jobs: List[Job] = field(init=False)

    def __post_init__(self) -> None:
        parsed = yaml.safe_load(open(self.filepath, 'r').read())
        assert 'parameters' in parsed.keys(), 'Missing parameters'
        assert 'jobs' in parsed.keys(), 'Missing jobs'
        self.raw = parsed
        self.params = Params(**parsed['parameters'])
        self.jobs = [Job(**j) for j in parsed['jobs']]

    def parse_jobs(self) -> None:
        parsed_jobs = []
        for job in self.jobs:
            logging.debug(job.to_dict())
            job.parse(folder=self.params.folder, env=self.params.env)
            parsed_jobs.append(job)
        self.jobs = parsed_jobs

    def save_output(self) -> None:
        output = {
            'parameters': self.params.to_dict(),
            'jobs': [j.to_dict() for j in self.jobs]
        }
        _, ext = os.path.splitext(self.params.output)
        assert ext == '.json', f'Not a JSON file: {self.params.output}'
        with open(self.params.output, 'w') as f:
            json.dump(output, f, sort_keys=False)



from typing import List

from .structures import clauses as cl, basics as bs


def remove_comments(text: str) -> str:

    new_lines = []

    # loop through each line
    for line in text.split('\n'):

        # loop through each character
        for i in range(len(line)):

            # if exist comment characters, cut short the line
            if line[i] == '#':
                line = line[:i]
                break
            if i > 0 and line[i-1:i+1] == '--':
                line = line[:i-1]
                break

        # keep line if it's not a comment line
        if len(line.strip()) > 0:
            new_lines.append(line.strip())

    return '\n'.join(new_lines)


def split_with_brackets(text: str) -> List[str]:
    output = []
    tmp = ''
    counter = 0
    for char in text:
        if char == ',' and counter == 0:
            output.append(tmp.strip())
            tmp = ''
            continue
        if char == ',' and counter > 0:
            tmp += char
            continue
        if char == '(':
            counter += 1
            tmp += char
            continue
        if char == ')':
            counter -= 1
            tmp += char
            continue
        tmp += char

    if len(tmp) > 0:
        output.append(tmp)

    return output


def print_clauses(clause_list: List[cl.Clause], indent: int = 0) -> None:

    for clause in clause_list:

        if isinstance(clause, cl.CteEndClause) \
                or isinstance(clause, cl.SubqueryEndClause):
            indent -= 2

        print(f'{" "*indent}{clause}')

        if isinstance(clause, cl.CteStartClause) \
                or isinstance(clause, cl.SubqueryStartClause):
            indent += 2


def print_basics(basic_list: List[bs.BasicQuery]) -> None:
    indent = 0
    for basic in basic_list:
        print(f'{basic.__class__.__name__}(clauses=[')
        print_clauses(basic.clauses, indent+2)
        print(f'])')



from abc import ABC
from dataclasses import dataclass


@dataclass
class Entity(ABC):
    name: str = None
    alias: str = None


@dataclass
class Table(Entity):
    pass


@dataclass
class Column(Entity):
    pass


import uuid
from typing import List, Optional

from . import clauses as cl
from ..graph import entities as en, relationships as rs


class BasicQuery:

    def __init__(self) -> None:
        self.clauses: List[cl.Clause] = []
        self._target: Optional[str] = None

    @property
    def target(self) -> Optional[str]:
        return self._target or uuid.uuid4().hex[:5]

    @target.setter
    def target(self, target: str) -> None:
        self._target = target

    def add_clause(self, clause: cl.Clause) -> None:
        self.clauses.append(clause)

    def pending_clause(self) -> bool:
        cls = self.clauses
        no_select = not any([isinstance(c, cl.SelectClause) for c in cls])
        no_from = not any([isinstance(c, cl.FromClause) for c in cls])
        n_sub_start = sum([isinstance(c, cl.SubqueryStartClause) for c in cls])
        n_sub_end = sum([isinstance(c, cl.SubqueryEndClause) for c in cls])
        last_union = isinstance(cls[-1], cl.UnionClause)
        return no_select or no_from \
               or (n_sub_start > n_sub_end) \
               or last_union

    def extract_table_to_table(self) -> List[rs.TableToTable]:
        target = en.Table(name=self.target)
        output = []
        for clause in self.clauses:
            for source in clause.extract_tables():
                rship = rs.TableToTable(source=source, target=target)
                output.append(rship)
        return output


class CteBasicQuery(BasicQuery):

    @property
    def target(self) -> Optional[str]:
        if len(self.clauses) == 0:
            return None
        assert isinstance(self.clauses[0], cl.CteStartClause)
        return self.clauses[0].extract_target()

    def pending_clause(self) -> bool:
        has_start = any([type(c) == cl.CteStartClause for c in self.clauses])
        has_end = any([type(c) == cl.CteEndClause for c in self.clauses])
        return has_start and not has_end

    def extract_table_to_table(self) -> List[rs.TableToTable]:
        assert isinstance(self.clauses[0], cl.CteStartClause)
        target = en.Table(name=self.target)
        output = []
        for clause in self.clauses[1:]:
            for source in clause.extract_tables():
                rship = rs.TableToTable(source=source, target=target)
                output.append(rship)
        return output




from pathlib import Path
from setuptools import setup, find_packages

PROJECT_DIR = Path(__file__).parent

long_description = (PROJECT_DIR / 'README.md').read_text()

setup(
    name='sqlexpress',
    version='0.2.1',
    author='sssiong',
    description='Package to parse SQL scripts for information',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sssiong/sqlexpress',
    include_package_data=True,
    install_requires=[
        'Flask==2.1.2',
        'networkx==2.8',
        'PyYAML==6.0',
    ],
    tests_require=['pytest'],
    packages=find_packages(exclude=('tests', )),
)